/*	motion.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"
#include "conf.h"

extern struct config conf;

int deamon=0;
int snapshot=0;
int alarmtime=0;
int fmt=VIDEO_PALETTE_RGB24;
int read_img=0;

char *start_capture (int dev, int width, int height, int input, int norm)
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
	struct video_window vid_win;
	struct video_mbuf vid_buf;
	char *map;

	if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (NULL);
	}
	if (vid_caps.type & VID_TYPE_MONOCHROME) fmt=VIDEO_PALETTE_GREY;
	/* Set channels and norms, NOT TESTED my philips cam doesn't have a
	 * tuner. update: it works with my brothers haupage wintv (bttv driver)
	 */
        if (input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev, VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = input;
                        vid_chnl.norm    = norm;
                        if (ioctl (dev, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (NULL);
                        }
                }
        }
	if (ioctl (dev, VIDIOCGMBUF, &vid_buf) == -1) {
		fprintf(stderr, "no mmap falling back on read\n");
		if (ioctl (dev, VIDIOCGWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCGWIN");
			return (NULL);
		}
		vid_win.width=width;
		vid_win.height=height;
		if (ioctl (dev, VIDIOCSWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCSWIN");
			return (NULL);
		}
		read_img=1;
		map=malloc(width*height*3);
		return (map);
	}
	/* If we are going to capture greyscale we need room to blow the image up */
	if (fmt==VIDEO_PALETTE_GREY)
		map=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	else
		map=mmap(0, vid_buf.size, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	
	if ((unsigned char *)-1 == (unsigned char *)map)
		return (NULL);
	return map;
}

char *next_capture (int dev, char *map, int width, int height)
{
	int i;
	struct video_mmap vid_mmap;

    	sigset_t    set, old;

	if (read_img) {
		if (fmt==VIDEO_PALETTE_GREY) {
			if (read(dev, map, width*height) != width*height)
				return NULL;
		} else {
			if (read(dev, map, width*height*3) != width*height*3)
				return NULL;
		}
	} else {
		vid_mmap.format=fmt;
		vid_mmap.frame=0;
		vid_mmap.width=width;
		vid_mmap.height=height;
    
        	sigemptyset (&set);	     //BTTV hates signals during IOCTL
	        sigaddset (&set, SIGCHLD);   //block SIGCHLD & SIGALRM
       		sigaddset (&set, SIGALRM);   //for the time of ioctls
        	sigprocmask (SIG_BLOCK, &set, &old);
                    
		if (ioctl(dev, VIDIOCMCAPTURE, &vid_mmap) == -1) {
	        	sigprocmask (SIG_UNBLOCK, &old, &set);
			return (NULL);
		}
		if (ioctl(dev, VIDIOCSYNC, &vid_mmap) == -1) {
	        	sigprocmask (SIG_UNBLOCK, &old, &set);
			return (NULL);
		}
		
        	sigprocmask (SIG_UNBLOCK, &old, &set); //undo the signal blocking
	}
	/* Blow up a grey */
	if (fmt==VIDEO_PALETTE_GREY) {
		for (i=width*height*3; i>=0; i--) {
			map[i]=map[i/3];
		}
	}
	return map;
}

void put_image_jpeg (FILE *picture, char *image, int width, int height, int quality)
{
        int y, x, line_width;
        JSAMPROW row_ptr[1];
        struct jpeg_compress_struct cjpeg;
        struct jpeg_error_mgr jerr;
        unsigned char *line;

        line = malloc (width * 3);
        if (!line)
                return;
        cjpeg.err = jpeg_std_error(&jerr);
        jpeg_create_compress (&cjpeg);
        cjpeg.image_width = width;
        cjpeg.image_height= height;
        cjpeg.input_components = 3;
        cjpeg.in_color_space = JCS_RGB;
        jpeg_set_defaults (&cjpeg);

        jpeg_set_quality (&cjpeg, quality, TRUE);
        cjpeg.dct_method = JDCT_FASTEST;
        jpeg_stdio_dest (&cjpeg, picture);

        jpeg_start_compress (&cjpeg, TRUE);

        row_ptr[0] = line;
        line_width = width * 3;
	for ( y = 0; y < height; y++) {
        for (x = 0; x < line_width; x+=3) {
                        line[x]   = image[x+2];
                        line[x+1] = image[x+1];
                        line[x+2] = image[x];
                }
                jpeg_write_scanlines (&cjpeg, row_ptr, 1);
                image += line_width;
        }
        jpeg_finish_compress (&cjpeg);
        jpeg_destroy_compress (&cjpeg);
        free (line);
}

void send_sms(char *sms_nr, struct tm *currenttime)
{
	char message[150];

	if (!fork()) {
	    	printf("Sending SMS message to %s\n", sms_nr);
		sprintf(message, "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		execlp("sms_client", "sms_client", sms_nr, message, 0);
		perror("Unable to send SMS message");
		exit(0);
	}
	return;
}

void send_mail(char *mail_address, struct tm *currenttime)
{
	char message[150];
	int fdpipe[2];
	
	if (!fork()) {
	    	printf("Sending mail to %s\n", mail_address);
		sprintf(message, "This is an automated message generated by motion\n\n"
		    "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n"
		    ".\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		pipe(fdpipe);
		if (!deamon) {
			close(0);
			dup(fdpipe[0]);
			close(fdpipe[0]);
		}
		write(fdpipe[1], message, strlen(message));
		close(fdpipe[1]);
		execlp("mail", "mail", mail_address, "-s", "MOTION ALERT", 0);
		perror("Unable to send message");
		exit(0);
	}
	return;
}

void exec_externcommand (char *externcommand)
{
	if (!fork()) {
		printf("Executing external command '%s'\n", externcommand);
		execlp(externcommand, externcommand, 0);
		perror("Unable to start external command");
		exit(-1);
	}
	return;
}

void create_mpeg (int event_nr, int filesc, char **files)
{
	int i;
	char params[255];
	char eventname[255];
	char *mpegname=eventname;
	FILE *paramfile=NULL;

	if (!filesc) return;
	strncpy(eventname, files[0], strlen(files[0])-7);
	if (strlen(files[0])==25) {
		eventname[strlen(files[0])-7]='m';
		eventname[strlen(files[0])-8]=0;
	} else {
		eventname[strlen(files[0])-7]=0;
	}
	/* Cut of eventnr, we have te date so we don't need to sort */
	while((mpegname++)[0]!='-');
	sprintf(params, "%s.params", mpegname);
	paramfile=fopen(params, "w");
	if (!fork()) {
		printf("Starting mpeg encoding of event %d\n", event_nr);
		fprintf(paramfile,
			"PATTERN IBBPBBPBBPBBPBBP\n"
			"OUTPUT %s.mpg\n"
			"BASE_FILE_FORMAT JPEG\n"
			"INPUT_CONVERT *\n"
			"GOP_SIZE 30\n"
			"SLICES_PER_FRAME 1\n"
			"INPUT_DIR .\n"
			"INPUT\n",
			mpegname);
		for (i=0; i<filesc; i++) fprintf(paramfile, "%s\n", files[i]);
		fprintf(paramfile,
			"END_INPUT\n"
			"PIXEL HALF\n"
			"RANGE 10\n"
			"PSEARCH_ALG LOGARITHMIC\n"
			"BSEARCH_ALG CROSS2\n"
			"IQSCALE 8\n"
			"PQSCALE 10\n"
			"BQSCALE 25\n"
			"REFERENCE_FRAME DECODED\n"
			"FRAME_RATE 25\n");
		fclose(paramfile);
		if (!fork()) {
			nice(5);
			execlp("mpeg_encode", "mpeg_encode", params, 0);
			perror("Failed to start mpeg_encode");
			exit(-1);
		}
		if (wait(0)==-1) exit(-1);
		for (i=0; i<filesc; i++) {
			remove(files[i]);
			free(files[i]);
		}
		free(files);
		remove(params);
		exit(0);
	}
	return;
}

void locate_diffs (char *out, char *new, int width, int height)
{
	int x, y, maxx=0, maxy=0, minx=width, miny=height;

	/* Locate movement */	
	for (x=0; x<width; x++) {
		for (y=0; y<height; y++) {
			if (out[(x+y*width)*3+0] && out[(x+y*width)*3+1] && out[(x+y*width)*3+2]) {
				if (x < minx) minx=x;
				if (x > maxx) maxx=x;
				if (y < miny) miny=y;
				if (y > maxy) maxy=y;
			}
		}
	}
	/* Draw a box around the movement */
	for (x=minx; x<=maxx; x++) {
		new[(x+width*miny)*3+0]=255;
		new[(x+width*miny)*3+1]=255;
		new[(x+width*miny)*3+2]=255;
	}
	for (x=minx; x<=maxx; x++) {
		new[(x+width*maxy)*3+0]=255;
		new[(x+width*maxy)*3+1]=255;
		new[(x+width*maxy)*3+2]=255;
	}
	for (y=miny; y<=maxy; y++) {
		new[(minx+y*width)*3+0]=255;
		new[(minx+y*width)*3+1]=255;
		new[(minx+y*width)*3+2]=255;
	}
	for (y=miny; y<=maxy; y++) {
		new[(maxx+y*width)*3+0]=255;
		new[(maxx+y*width)*3+1]=255;
		new[(maxx+y*width)*3+2]=255;
	}
}

void sig_alarm (int signo) 
{
	/* Somebody (maybe we ourself) wants us to make a snapshot */
	snapshot=1;
	/* Set everything for the next snapshot */
	signal(SIGALRM, sig_alarm);
	if (alarmtime) alarm(alarmtime);
	return;
}

int diff_standard (char *image_ref, char *image_new, char *image_out, int width, int height)
{
	int i, diffs=0;

	for (i=0; i<width*height*3; i++) {
		image_out[i]=image_ref[i]-image_new[i];
		if ((signed char)image_out[i] > 32 || (signed char)image_out[i] < -32) {
			image_out[i]=image_new[i];
			diffs++;
		} else {
			image_out[i]=0;
		}
	}
	return diffs;
}

int main (int argc, char **argv)
{
	int dev, i, diffs, shots=0, event_nr=0, prev_event=0, filesc=0, filesmc=0;
	unsigned char *image_ref, *image_new, *image_out;
	time_t currenttimep, lasttime=0;
	struct tm *currenttime=NULL;
	char file[255], filebase[255];
	char **files=NULL, **filesm=NULL;
	FILE *picture;

	conf_load(argc, argv);
	
	if (conf.filepath) if (chdir(conf.filepath)) {
		perror("Unable to chdir to target directory");
		exit(-1);
	}

	/* Set signal handlers */
	signal(SIGCHLD, SIG_IGN);
	signal(SIGALRM, sig_alarm);

	image_ref=malloc(conf.width*conf.height*3);
	image_out=malloc(conf.width*conf.height*3);

	dev=open (conf.device, O_RDWR);
	if (dev < 0) {
		perror ("Failed to open video device");
		exit(-1);
	}

	/* Deamonize */
	if (deamon) {
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}
	/* set the device settings */
	image_new=start_capture (dev, conf.width, conf.height, conf.input, conf.norm);
	if (!image_new) {
		perror("Capture error");
		exit(-1);
	}
	/* Allow videodevice to settle in */
	sleep(1);
	/* Capture first image, or we will get an alarm on start */
	if (!next_capture(dev, image_new, conf.width, conf.height)) {
		perror("Capture error");
		exit(-1);
	}
	memcpy(image_ref, image_new, conf.width*conf.height*3);

	if (alarmtime) alarm(alarmtime);

	/* Should go on forever... unless you bought vaporware :) */
	while (next_capture(dev, image_new, conf.width, conf.height)) {
		/* Make a differences picture in image_out */
		diffs=diff_standard(image_ref, image_new, image_out, conf.width, conf.height);
		for (i=0; i<conf.width*conf.height*3; i++) {
			/* Old image slowly decays, this will make it even harder on
			 * slow moving object to stay undetected
			 */
			image_ref[i]=(image_ref[i]+image_new[i])/2;
		}
		currenttimep=time(NULL);
		currenttime=localtime(&currenttimep);
		if (lasttime!=currenttimep) {
			shots=-1;
		}
		if (currenttimep-lasttime>= conf.gap && lasttime) {
			if (event_nr==prev_event) {
				event_nr++;
    				if (conf.mpeg) {
					create_mpeg(prev_event, filesc, files);
					create_mpeg(prev_event, filesmc, filesm);
				} else {
					for (i=0; i<filesc; i++) free(files[i]);
					free(files);
					for (i=0; i<filesmc; i++) free(filesm[i]);
					free(filesm);
				}
				filesc=0;
				files=NULL;
				filesmc=0;
				filesm=NULL;
			}
		}
		/* The cat just walked in :) */
		if (diffs > conf.max_changes) {
			printf("\007Motion detected %d\n", diffs);
			/* Take action if necesarry */
			if (event_nr!=prev_event) {
				if (conf.sms_nr) send_sms(conf.sms_nr, currenttime);
				if (conf.mail_address) send_mail(conf.mail_address, currenttime);
				if (conf.externcommand) exec_externcommand(conf.externcommand);
				prev_event=event_nr;
			}
			lasttime=currenttimep;
		} else if (conf.always_diff) {
			printf("changes: %d\n", diffs);
		}
		if (diffs > conf.max_changes && shots<conf.frame_limit-1) {
			shots++;
			/* Output the latest picture 'image_new' or image_out for motion picture.
			 */
			sprintf(filebase, "%02d-%04d%02d%02d%02d%02d%02d-%02d",
			    event_nr,
			    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
			    currenttime->tm_mday, currenttime->tm_hour,
			    currenttime->tm_min, currenttime->tm_sec, shots);
			if (conf.motion_img) {
				sprintf(file, "%sm.jpg", filebase);
				printf("Motion picture saved to: %s\n", file);
				picture=fopen(file, "w");
				filesmc++;
				filesm=realloc(filesm, filesmc*(sizeof(void *)));
				filesm[filesmc-1]=malloc(30);
				strncpy(filesm[filesmc-1], file, 30);
				if (conf.locate && diffs > conf.max_changes)
					locate_diffs(image_out, image_out, conf.width, conf.height);
				put_image_jpeg (picture, image_out, conf.width, conf.height, conf.quality);
				fclose(picture);
			} 
			if (conf.new_img) {
				sprintf(file, "%s.jpg", filebase);
				printf("Normal picture saved to: %s\n", file);
				filesc++;
				files=realloc(files, filesc*(sizeof(void *)));
				files[filesc-1]=malloc(30);
				strncpy(files[filesc-1], file, 30);
				picture=fopen(file, "w");
				if (conf.locate && diffs > conf.max_changes)
					locate_diffs(image_out, image_new, conf.width, conf.height);
				put_image_jpeg (picture, image_new, conf.width, conf.height, conf.quality);
				fclose(picture);
			}
		}
		if (snapshot) {
			sprintf(filebase, "%04d%02d%02d%02d%02d%02d-snapshot",
			    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
			    currenttime->tm_mday, currenttime->tm_hour,
			    currenttime->tm_min, currenttime->tm_sec);
			sprintf(file, "%s.jpg", filebase);
			printf("Snapshot saved to: %s\n", file);
			picture=fopen(file, "w");
			put_image_jpeg (picture, image_new, conf.width, conf.height, conf.quality);
			fclose(picture);
			snapshot=0;
		}
		fflush(0);
	}
	perror("Somebody stole the video device, lets hope we got his picture");
	close (dev);
	free(image_ref);
	free(image_out);
	exit(0);
}
